(*
 Oukseh Lee, The LET Project, KAIST

 This source is almost copied from the Objective Caml compiler.
*)



(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)


structure NTmain = struct

open Config
open Clflags

structure NTcompile = CompileGen(Compile)

fun process_interface_file ppf name =
  Compile.interface ppf name

fun process_implementation_file ppf name = (
  Compile.init_path();
  Compile.implementation ppf Compile.initial_env Parse.implementation 
     String.capitalize name;
  objfiles := (Filename.chop_extension name ^ ".cmo") :: !objfiles)

fun process_file ppf name =
  if Filename.check_suffix name ".ml"
  || Filename.check_suffix name ".mlt" then 
    process_implementation_file ppf name
  else if Filename.check_suffix name !Config.interface_suffix then
    process_interface_file ppf name
(***** poisson *****)
  else if Filename.check_suffix name ".n"
       || Filename.check_suffix name ".nt" then (
    NTcompile.implementation ppf name;
    objfiles := (Filename.chop_extension name ^ ".cmo") :: !objfiles;
    ntyfiles := (Filename.chop_extension name ^ ".nty") :: !ntyfiles
  )
  else if Filename.check_suffix name ".ns" then (
    NTcompile.compile_signature ppf name;
    ntyfiles := (Filename.chop_extension name ^ ".nty") :: !ntyfiles
  )
  else if Filename.check_suffix name ".nty" then
    ntyfiles := name :: !ntyfiles
(*******************)
  else if Filename.check_suffix name ".cmo"
       || Filename.check_suffix name ".cma" then
    objfiles := name :: !objfiles
  else if Filename.check_suffix name ext_obj
       || Filename.check_suffix name ext_lib then
    ccobjs := name :: !ccobjs
  else if Filename.check_suffix name ext_dll then
    dllibs := name :: !dllibs
  else if Filename.check_suffix name ".c" then (
    Compile.c_file name;
    case Sys.os_type of
      "MacOS" => ccobjs := (name ^ ".o") :: (name ^ ".x") :: !ccobjs
    | _ =>
       ccobjs := (Filename.chop_suffix (Filename.basename name) ".c" ^ ext_obj)
                 :: !ccobjs
  )
  else
    raise(Arg.Bad("don't know what to do with " ^ name))

fun anonymous x = process_file Format.err_formatter x

fun print_version_number() = (
  Nconfig.print_header();
  print_string "Standard library directory: ";
  print_string Config.standard_library; print_newline();
  exit 0)

val usage = "Usage: nmlc <options> <files>\nOptions are:"

open Noption 

structure Options = OptionsFn (struct
  val anonymous = anonymous 
  fun impl x = process_implementation_file Format.err_formatter x
  fun intf x = process_interface_file Format.err_formatter x
  val print_version_number = print_version_number
end)

fun main () = (
    Arg.parse Options.list anonymous usage;
    if !make_archive then 
       if !objfiles = [] then NTcompile.create_archive !ntyfiles !archive_name
       else (
          Compile.init_path();
          Bytelibrarian.create_archive (List.rev !objfiles) !archive_name
       )
    else if not !compile_only && !objfiles <> [] then (
      Compile.init_path();
      Bytelink.link (List.rev !objfiles)
    );
    exit 0
  ) handle x => (
    NTerrors.report_error Format.std_formatter Errors.report_error x; 
    exit 2
  )
end

val _ = Printexc.catch NTmain.main ()



         
