(*
 Oukseh Lee, The LET Project, KAIST

 This source is almost copied from the Objective Caml compiler.
*)

(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ntopmain.n,v 1.1.1.1 2002/01/16 09:01:31 cookcu Exp $ *)

(* parse_use_file, use_file, use_silently, loop, run_script 
   are imported from ntoploop.ml  -- cookcu -- 
*)

structure Ntopmain = struct
open Lexing
open Format
open Config
open Misc
open Parsetree
open Types
open Typedtree
open N2caml
open Clflags
open Toploop

structure NTcompile = CompileGen(Compile)

(*

let parse_toplevel_phrase = ref Parse.toplevel_phrase
*)
(*********** poisson ***********)

structure S = String_ast.Ast
structure A = Fake.Ast

local
 val sloc = Location.rhs_loc
 val nloc = Location.none
in
 fun mktopexp e =
        S.Str(
          S.SimpleDec(
            S.ValDec([],
              [(S.Nonrec, S.VarPat(("it",nloc),nloc),e,nloc)],
              nloc),
            nloc),
          nloc)
end

fun rmtopexp (S.Str(S.SimpleDec(S.ValDec(_,[(_,_,S.SeqExp(l,_),_)],_),_),_)) = l
  | rmtopexp (S.Str(S.SimpleDec(S.ValDec(_,[(_,_,e,_)],_),_),_)) = [e]
  | rmtopexp _ = raise TyDebug.Bug "Topmain.rmtopexp"

fun protect r newval body =
  let val oldval = !r 
  in  (r := newval;
       let val res = body() 
       in  r := oldval;
           res
       end) handle x => (r := oldval; raise x)
  end

fun typingit td = if not (!Clflags.ntyping) then td else
  let val i0 = Infer.infer td
      val i1 = Remove_record.topdec i0 
      val i2 = Absyn2Ast.topdec i1
  in  
      NEnv.stabilize_top();
(*      EnvPrint.b(NEnv.get_b()); *)
      i2
  end

fun parse_toplevel_phr lb =
  let val nast = Nparse.toplevel_parse lb
      val nast' = case nast of 
                    S.TopDec(d,i) => S.TopDec(typingit d, i) 
                  | S.TopExp(e,i) => S.TopDec(typingit (mktopexp e), i)
                  | x => x
      val nast' = (Nconvert.convert_toplevel(Ast2Fake.toplevel nast')) 
  in  nast'
  end

fun parse_usefile lb =
  let val nast = Nparse.implementation_parse lb
  in  [A.TopDef (Nconvert.convert_topdec (Ast2Fake.topdec (typingit nast)))]
  end


fun print_result_value env sg n =
  let fun find (Tsig_value(id,decl)::rem) = 
            if n=Ident.name id then (
              printf "@ = @[";
              Toploop.print_value env 
                                  (getvalue (Translmod.toplevel_name id))
                                  std_formatter 
                                  decl.val_type;
              printf "@]")
            else find rem
        | find (_::rem) = find rem
        | find _ = printf " = unbound value"
  in  find sg
  end

   
fun execute_phrase print_outcome ppf source =
  let val phr = trans_toplevel source in
  if !Clflags.dump_parsetree then Printast.top_phrase ppf phr;
  case (phr, source) of
    (Ptop_def sstr, A.TopDef nast) =>
      let val (str, sg, newenv) = Typemod.type_structure !toplevel_env sstr
          val lam = Translmod.transl_toplevel_definition str
          val _ = Warnings.check_fatal ()
          val res = load_lambda ppf lam 
      in  case res of
            Result v =>
              if print_outcome then 
                (EnvPrint.b (!NEnv.s2) (print_result_value newenv sg);
                 NEnv.move_b_to_base());
              toplevel_env := newenv;
              true
          | Exception excn => print_exception_outcome ppf excn; false
      end
  | (Ptop_dir(dir_name, dir_arg),_) =>
      ((case (Hashtbl.find directive_table dir_name, dir_arg) of
          (Directive_none f, Pdir_none) => f (); true
        | (Directive_string f, Pdir_string s) => f s; true
        | (Directive_int f, Pdir_int n) => f n; true
        | (Directive_ident f, Pdir_ident lid) => f lid; true
        | (Directive_bool f, Pdir_bool b) => f b; true
        | (_, _) =>
            fprintf ppf "Wrong type of argument for directive `%s'.@." dir_name;
            false
      ) handle Not_found => (
        fprintf ppf "Unknown directive `%s'.@." dir_name;
        false))
  | _ => raise TyDebug.Bug "Ntopmain.execute_phrase"
 end

fun use_file ppf name = 
  let val filename = find_in_path !Config.load_path name
      val ic = open_in_bin filename
      val lb = Lexing.from_channel ic
    (* Skip initial #! line if any *)
      val buffer = String.create 2
      val _ = if input ic buffer 0 2 = 2 && buffer = "#!"
              then ignore(input_line ic)
              else seek_in ic 0
      
      val success =
            protect Location.input_name filename (fn () =>
             ((List.iter
                (fn ph => 
                   if not (execute_phrase !use_print_results ppf ph) then raise Exit)
                (parse_usefile lb);
              true)
              handle
                Exit => false
              | Sys.Break => fprintf ppf "Interrupted.@."; false
              | x => NTerrors.report_error ppf Errors.report_error x; false)) 
  in  close_in ic;
      success
  end handle Not_found => fprintf ppf "Cannot find file %s.@." name; false

fun use_silently ppf name =
  protect use_print_results false (fn () => use_file ppf name)


fun dir_use ppf name = ignore(use_file ppf name)

val _ = Hashtbl.add directive_table "use" (Directive_string (dir_use std_formatter))


(* Toplevel initialization. Performed here instead of at the
   beginning of loop() so that user code linked in with ocamlmktop
   can call directives from Topdirs. *)

val _ = (
  Sys.interactive := true;
  Symtable.init_toplevel();
  Compile.init_path())

fun load_ocamlinit ppf =
  if Sys.file_exists ".nmlinit" then ignore(use_silently ppf ".nmlinit")

(* The interactive loop *)

exception PPerror

fun loop ppf = (
  Nconfig.print_full_header();
  (* Add whatever -I options have been specified on the command line,
     but keep the directories that user code linked in with ocamlmktop
     may have added to load_path. *)
(*******************)
  load_path := "" :: (List.rev !Clflags.include_dirs @ !load_path);
  Dll.add_path !load_path;
  toplevel_env := NTcompile.initial_env();
  let val lb = Lexing.from_function refill_lexbuf 
  in  Location.input_name := "";
      Location.input_lexbuf := Some lb;
      Sys.catch_break true;
      load_ocamlinit ppf;
      while true do (
        empty_lexbuf lb;
        Location.reset();
        first_line := true;
        NEnv.reset_top();
        let val phr = (parse_toplevel_phr lb) handle Exit => raise PPerror 
        in  ignore(execute_phrase true ppf phr)
        end) handle End_of_file => exit 0
                  | Sys.Break => fprintf ppf "Interrupted.@."
                  | PPerror => ()
                  | x => NTerrors.report_error ppf Errors.report_error x
      end
  end)

(* Execute a script *)

fun run_script ppf name args =
  let fun find n = (
        if n >= Array.length args then invalid_arg "Ntopmain.run_script";
        if args.[n] = name then n else find (n+1))
      val pos = find 0 
      val len = Array.length args - pos
  in  if Array.length Sys.argv < len then invalid_arg "Ntopmain.run_script";
      Array.blit args pos Sys.argv 0 len;
      Obj.truncate (Obj.repr Sys.argv) len;
      Arg.current := 0;
      Compile.init_path();
      toplevel_env := NTcompile.initial_env();
      use_silently ppf name
  end

val usage = "Usage: nml <options> [script-file]\noptions are:"

fun file_argument name =
  exit (if run_script Format.err_formatter name Sys.argv then 0 else 2)

fun main () = (
  Arg.parse [
     ("-I", Arg.String(fn dir => 
         let val dir = Misc.expand_directory Config.standard_library dir
         in  include_dirs := dir :: !include_dirs
         end),
           "<dir>  Add <dir> to the list of include directories"),
     ("-labels", Arg.Clear classic, " Use commuting label mode"),
     ("-noassert", Arg.Set noassert, " Do not compile assertion checks"),
     ("-rectypes", Arg.Set recursive_types, " Allow arbitrary recursive types"),
     ("-unsafe", Arg.Set fast, " No bound checking on array and string access"),
     ("-w", Arg.String (Warnings.parse_options false),
           "<flags>  Enable or disable warnings according to <flags>:\n\
       \032    A/a enable/disable all warnings\n\
       \032    C/c enable/disable suspicious comment\n\
       \032    F/f enable/disable partially applied function\n\
       \032    M/m enable/disable overriden method\n\
       \032    P/p enable/disable partial match\n\
       \032    S/s enable/disable non-unit statement\n\
       \032    U/u enable/disable unused match case\n\
       \032    V/v enable/disable hidden instance variable\n\
       \032    X/x enable/disable all other warnings\n\
       \032    default setting is A (all warnings enabled)"),
     ("-warn-error" , Arg.String (Warnings.parse_options true),
       "<flags>  Enable or disable fatal warnings according to <flags>\n\
         \032    (see option -w for the list of flags)\n\
         \032    default setting is a (all warnings are non-fatal)"),
     ("-dparsetree", Arg.Set dump_parsetree, " (undocumented)"),
     ("-drawlambda", Arg.Set dump_rawlambda, " (undocumented)"),
     ("-dlambda", Arg.Set dump_lambda, " (undocumented)"),
     ("-dinstr", Arg.Set dump_instr, " (undocumented)"),
     ("-ntyping", Arg.Set ntyping, " turn on the nML type-checker"),
     ("-nontyping", Arg.Clear ntyping, " turn off the nML type-checker") 
    ] file_argument usage;
  loop Format.std_formatter)

end

val _ = Printexc.catch Ntopmain.main ()
