(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

structure Ast2Fake = struct
  structure I = String_ast.Ast
  structure A = Fake.Ast

  val count = ref 0
  fun new_variable loc = 
    let val c = !count
    in  count := !count + 1;
        ("@ast2fask"^string_of_int(c),loc)
    end

  val map = List.map
  fun typ (t,_) = ty t

  and con (i,Some t,a) = (i,Some(ty t),a)
    | con (i,None,a) = (i,None,a)

  and ty t = case t of
      I.VarTy(tv,a) => A.VarTy(tv,a)
    | I.ConstTy(l,id,a) => A.ConstTy(List.map ty l, id, a)
    | I.RecordTy(l,a) => A.RecordTy(List.map (fn (l,t,b) => (l, ty t, b)) l, a)
    | I.TupleTy(l,a) => A.TupleTy(List.map ty l, a)
    | I.FunTy(t1,t2,a) => A.FunTy(ty t1, ty t2, a)

  fun pat p = case p of
        I.WildPat a => A.WildPat a
      | I.UnitPat a => A.UnitPat a
      | I.IntPat(i,a) => A.IntPat(i,a)
      | I.StringPat(s,a) => A.StringPat(s,a)
      | I.CharPat(c,a) => A.CharPat(c,a)
      | I.VarPat(i,a) => A.VarPat(i,a)
      | I.ConPat(i,a) => A.ConPat(i,a)
      | I.AppPat(i,p,a) => A.AppPat(i,pat p,a)
      | I.RecordPat(l,a) => A.RecordPat(map labpat l,a)
      | I.SubRecordPat(l,a) => A.SubRecordPat(map labpat l,a)
      | I.RefPat(p,a) => A.RefPat(pat p,a)
      | I.AsPat(i,Some t,p,a) => A.AsPat(i,Some(ty t),pat p,a)
      | I.AsPat(i,None,p,a) => A.AsPat(i,None,pat p, a)
      | I.TuplePat(l,a) => A.TuplePat(map pat l,a)
      | I.ListPat(l,a) => A.ListPat(map pat l,a)
      | I.ArrayPat(l,a) => A.ArrayPat(map pat l,a)
      | I.OrPat(l,a) => A.OrPat(map pat l,a)
      | I.ConstraintPat(p,t,a) => A.ConstraintPat(pat p,ty t,a)

  and labpat(l,p,a) = (l,pat p,a)
  
  fun make_app a l = case l of
        [] => raise (TyDebug.Bug "ast2fake.make_app")
      | [e] => e
      | h::t => A.AppExp(make_app a t, h, a)

  fun rec_or_not l = case l of
        [] => raise (TyDebug.Bug "ast2fake.rec_or_not")
      | [(b,_,_,_)] => b = I.Rec
      | (b,_,_,_)::t => 
          let val z = b = I.Rec
          in  if rec_or_not t = z then z else raise (TyDebug.Bug "ast2fake.rec_or_not2")
          end 

  fun exp e = case e of
        I.UnitExp a => A.UnitExp a
      | I.IntExp(i,a) => A.IntExp(i,a)
      | I.RealExp(s,a) => A.RealExp(s,a)
      | I.StringExp(s,a) => A.StringExp(s,a)
      | I.CharExp(c,a) => A.CharExp(c,a)
      | I.VarExp(i,a) => A.VarExp(i,a)
      | I.ConExp(i,a) => A.ConExp(i,a)
      | I.AppExp(e,l,d) => make_app d (List.rev (map exp (e::l)))
      | I.RecordExp(l,a) => A.RecordExp(map (fn(l,e,a)=>(l,exp e,a)) l,a)
      | I.RecordFieldExp(e,lab,a) => A.RecordFieldExp(exp e,lab,a)
      | I.ArrayFieldExp(e1,e2,a) => A.ArrayFieldExp(exp e1,exp e2,a)
      | I.UpdateArrayExp(e1,e2,e3,a) => A.UpdateArrayExp(exp e1,exp e2,exp e3,a)
      | I.SubstRecordExp(e1,l,e2,a) => A.SubstRecordExp(exp e1,l,exp e2,a)
      | I.TupleExp(l,a) => A.TupleExp(map exp l,a)
      | I.ListExp(l,a) => A.ListExp(map exp l,a)
      | I.ArrayExp(l,a) => A.ArrayExp(map exp l,a)
      | I.SeqExp(l,a) => A.SeqExp(map exp l,a)
      | I.LetExp(d,e,a) => A.LetExp(dec d,exp e,a)
      | I.HandleExp(e,l,a) => A.HandleExp(exp e,map rule l,a)
      | I.RaiseExp(e,a) => A.RaiseExp(exp e,a)
      | I.FnExp(l,a) => A.FnExp(map fnrule l,a)
      | I.AssignExp(e1,e2,a) => A.AssignExp(exp e1,exp e2,a)
      | I.RefExp(e,a) => A.RefExp(exp e,a)
      | I.DeRefExp(e,a) => A.DeRefExp(exp e,a)
      | I.CaseExp(e,l,a) => A.CaseExp(exp e,map rule l,a)
      | I.IfExp(e1,e2,e3,a) => A.IfExp(exp e1,exp e2,exp e3,a)
      | I.WhileExp(e1,e2,a) => A.WhileExp(exp e1,exp e2,a)
      | I.ForExp(i,e1,e2,e3,e4,a) => A.ForExp(i,exp e1,exp e2,exp e3,exp e4,a)
      | I.ConstraintExp(e,t,a) => A.ConstraintExp(exp e,ty t,a)

  and fnrule(l,e,a) = (map pat l,exp e,a)
  and rule(p,e,a) = (pat p,exp e,a)

  and dec d = case d of
        I.ValDec(l,l',a) => 
          if rec_or_not l' then
            A.RecValDec(l,map valbind l',a)
          else
            A.ValDec(l, map valbind l', a)
      | I.FunDec(l',l,a) => A.FunDec(l',map funbind l, a)
      | I.TypeDec(l,a) => A.TypeDec(map typebind l,a)
      | I.ExceptionDec(l,a) => A.ExceptionDec(map con l,a)
      | I.LocalDec(d,e,a) => A.LocalDec(dec d,dec e,a)
      | I.OpenDec(i,a) => A.OpenDec(i,a)
      | I.SeqDec(l,a) => A.SeqDec(map dec l,a)
  and valbind(r,p,e,a) = (pat p,exp e,a)
  and funbind(l,a) = (map (fn(i,l,e,a)=>(i,map pat l,exp e,a)) l,a)
  and typebind t = case t of
        I.TypeBind(l,i,t,a) => A.TypeBind(l,i,ty t,a)
      | I.DataBind(l,i,l',a) => A.DataBind(l,i,map con l',a)

  fun spec p = case p of
        I.ValSpec(l,a) => A.ValSpec(map (fn(i,t,a)=>(i,ty t,a)) l,a)
      | I.TypeSpec(l,a) => A.TypeSpec(map typedesc l,a)
      | I.ExnSpec(l,a) => A.ExnSpec(map con l,a)
      | I.IncludeSpec(g,a) => A.IncludeSpec(sigexp g,a)
      | I.StrSpec(l,a) => A.StrSpec(map (fn(i,g,a) =>(i,sigexp g,a)) l,a)
      | I.SeqSpec(l,a) => A.SeqSpec(map spec l,a)
  and typedesc d = case d of
        I.TypeDesc(l,i,a) => A.TypeDesc(l,i,a)
      | I.TypeBindDesc(l,i,t,a) => A.TypeBindDesc(l,i,ty t,a)
      | I.DataDesc(l,i,m,a) => A.DataDesc(l,i,map con m,a)

  and strexp e = case e of
        I.StrStr(d,a) => A.StrStr(strdec d,a)
      | I.VarStr(i,a) => A.VarStr(i,a)
      | I.SigStr(s,g,a) => A.SigStr(strexp s, sigexp g, a)
      | I.FctAppStr(i,l,a) => A.FctAppStr(i, map strexp l, a)
  and sigexp g = case g of
        I.VarSig(i,a) => A.VarSig(i,a)
      | I.SigSig(p,a) => A.SigSig(spec p,a)
      | I.ConstraintSig(g,(l,b),a) =>
          A.ConstraintSig(sigexp g,
                          (map (fn(l,i,t,a) => (l,i,ty t,a)) l,b),
                          a)
  and strdec d = case d of
        I.SimpleDec(d,a) => A.SimpleDec(dec d,a)
      | I.StrDec(l,a) => 
          A.StrDec(map (fn (i,None,s,a)=>(i,None,strexp s,a)
                         | (i,Some g,s,a) => (i,Some(sigexp g),strexp s,a)) l,
                   a)
      | I.SeqStrDec(l,a) => A.SeqStrDec(map strdec l,a)
 
  fun topdec p = case p of
        I.Sig((l,a),b) =>
          A.Sig((map (fn (a,b,c) => (a,sigexp b,c)) l,a),b)
      | I.Fct((i,l,None,s,a),loc) => 
          A.Fct((i, map (fn(a,b,c)=>(a,sigexp b,c)) l,None,strexp s,a), loc)
      | I.Fct((i,l,Some g,s,a),loc) => 
          A.Fct((i, map (fn(a,b,c)=>(a,sigexp b,c)) l,Some(sigexp g),strexp s,a), loc)
      | I.Str(d,a) => A.Str(strdec d,a)
      | I.SeqTopDec(l,a) => A.SeqTopDec(map topdec l,a)

  fun dirarg d = case d of
        I.NoneArg => A.ArgNone
      | I.StringArg s => A.ArgString s
      | I.LidentArg v => A.ArgLident v
      | I.IntArg i => A.ArgInt i

  fun toplevel l = case l of
        I.TopDec(d,i) => A.TopDef(topdec d)
      | I.TopExp(e,i) => A.TopExp([exp e], i)
      | I.TopDir(s,da,i) => A.TopDir(s, dirarg da)
end
