(*
 Hyunjun Eo, Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

open String_ast.Ast
open Strtbl

structure Strtbl_build = struct

fun rename "main" = "Main"
  | rename x = x

fun foreach_file fname dname self = 
  let val module_name = rename (Filename.chop_suffix 
                                 (Filename.basename fname) ".n")
      fun build_from_sigbind ((id,_),_,_) =
            if self then sig_add sig_table id "."
                    else sig_add sig_table id module_name
  
      fun build_from_strbind ((id,_),_,_,_) =
            if self then str_add str_table id "."
                    else str_add str_table id module_name
      fun build_from_sig (sblist,_) = List.iter build_from_sigbind sblist
      fun build_from_fct ((id,_),_,_,_,_) =
            if self then fct_add fct_table id "."
                    else fct_add fct_table id module_name
      fun build_from_str st = case st of
            SimpleDec (d,_) => ()
          | StrDec (sblist,_) => List.iter build_from_strbind sblist
          | SeqStrDec (sdlist,_) => List.iter build_from_str sdlist
      fun foreach_topdec ast = case ast of
            Sig (sg,_) => build_from_sig sg 
          | Fct (ft,_) => build_from_fct ft
          | Str (st,_) => build_from_str st
          | SeqTopDec (tdl,_) => List.iter foreach_topdec tdl
      val inputfile = if dname = "" then fname
                                    else Filename.concat dname fname 
      val ic = open_in_bin inputfile 
      val ast = (
            seek_in ic 0;
            Location.input_name := inputfile;
            Nparse.parse (Lexing.from_channel ic)
          ) handle x => (close_in ic; raise x)
  in  close_in ic;
      foreach_topdec ast
  end

(*
fun build_str_table sourcefile _ = 
  let val nfiles = open_in ".nfiles"
      val fname = ref ""
  in  while (fname := input_line nfiles; true) do
        if (Filename.check_suffix (!fname) ".n") then 
          ((foreach_file (!fname) "" 
                         (Filename.basename sourcefile = Filename.basename !fname)
           ) handle _ => ())
      end handle End_of_file => ()
               | _ => ()
  end handle Sys_error _ => ()
*)

fun build_str_table sourcefile load_path =
  let fun foreach_dir dname =
        if dname <> "" then 
          let val dh = Sys.opendir dname
              val fname = ref ""
          in  (while (fname := Sys.readdir dh; true) do
                if (Filename.check_suffix (!fname) ".n") then
                  ((foreach_file (!fname) dname
                                 (Filename.basename sourcefile = Filename.basename !fname)
                   ) handle _ => ())
               end handle End_of_file => ());
              Sys.closedir dh
          end handle _ => ()
  in  List.iter foreach_dir ((List.rev load_path) @ ["."])
  end
end
