(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)

(* $Id: location.mli,v 1.1.1.1 2002/01/16 09:01:30 cookcu Exp $ *)

(* Source code locations, used in parsetree *)
open Format

type t =
  { loc_start: int; loc_end: int; loc_ghost: bool }

val none: t
val symbol_rloc: unit -> t
val symbol_gloc: unit -> t
val rhs_loc: int -> t

val input_name: string ref
val input_lexbuf: Lexing.lexbuf option ref

val print: formatter -> t -> unit
val print_warning: t -> formatter -> Warnings.t -> unit
val prerr_warning: t -> Warnings.t -> unit
val echo_eof: unit -> unit
val reset: unit -> unit

val highlight_locations: t -> t -> bool

type order = LESS | GREATER | EQUAL | SUBSET | SUPSET | UNKNOWN
val loc_compare : t -> t -> order
val location_to_location : int * int * bool -> t
val location_field_loc_start : t -> int
val location_field_loc_end : t -> int
val location_field_loc_ghost : t -> bool

val incr_num_loc_lines : int -> unit
