(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)


structure TyError = struct
  open Info.StringInfo
  open Format
 structure A = String_ast.Ast

  type class = Var | Typ | Tv | Str | Sig | Fct
  type id = string
  type lid = id list
  type gid = class * lid
(*
  type id = Varid of A.varlongid
          | Tyid of A.tylongid
          | Tvid of tyvarinfo
          | Strid of A.strlongid
          | Sigid of sigidinfo
          | Fctid of fctidinfo
*)
  type match = 
         Domain of    gid
       | Value of     id * Ty.scheme * Ty.scheme
       | Type of      id * tymatch
  
  and tymatch = 
         Arity of     int * int
       | Unknown
       | Name of      Ty.Tn.t * Ty.Tn.t
       | String of    Ty.TS.t * Ty.TS.t
       | Constype of  id * (Ty.scheme option) * (Ty.scheme option)
       | Consexist of id
       | Consreq of   id
              
  type cases = SS | FI | FS

  type t = 
         Unbound of        gid
       | Duplicate of      gid
       | UnifyFail of      Ty.ty * Ty.ty
       | TsApply of        lid * int * int
       | Match of          cases * lid option * lid option * (id list * match)
       | RecursiveType of  id list
       | ConstraintSig of  lid
       | RConstraintSig of lid
       | FunName of        id * id
       | FunLength of      int * int

  type w =
         Overload
 
  exception TM of tymatch
  exception M of id list * match
  exception T of t * Location.t
  exception W of w * Location.t

  fun to_lid (a,(b,_),_) = (List.map (fn (x,_) => x) a)@[b]
  fun to_gid class x = (class, to_lid x)
  fun to_id (b,_) = b
  fun add_str str (x,y) = (x,str@y)

  fun unbound_var (id as (_,_,l)) = raise T(Unbound(to_gid Var id),l)
  fun unbound_ty (id as (_,_,l)) = raise T(Unbound(to_gid Typ id),l)
  fun unbound_str (id as (_,_,l)) = raise T(Unbound(to_gid Str id),l)
  fun unbound_tyvar (id,l) = raise T(Unbound(Tv,[id]),l)
  fun unbound_sig (id,l) = raise T(Unbound(Sig,[id]),l)
  fun unbound_fct (id,l) = raise T(Unbound(Fct,[id]),l)
  fun duplicate_var (id,l) = raise T(Duplicate(Var, [id]), l)
  fun ts_apply x y z l = raise T(TsApply(to_lid x,y,z),l)
  fun constraint_sig (id as (_,_,l)) = raise T(ConstraintSig(to_lid id),l)
  fun redundant_constraint_sig (id as (_,_,l)) = raise T(RConstraintSig(to_lid id),l)

  fun pp_error() = printf "@[<6>: "

  fun pp_scheme ((n,t):Ty.scheme) = (printf "@["; TyPrint.typ t; printf "@]")

  fun pp_list f g [] = ()
    | pp_list f g [h] = f h
    | pp_list f g (h::t) = (f h; g(); pp_list f g t)

  fun pp_lid lid = pp_list (printf "%s") (fn _ => printf ".") lid

  fun pp' (t,l) = let
    fun fv (class, x) = 
      let val hd = fn (Var | Typ) => "" | Str => " " 
                    | Sig => "Ÿ " | Fct => "Լ " | Tv => "Ÿ  "
          val tl = fn Typ => " Ÿ" | _ => " "
      in  printf "%s" (hd class);
          pp_lid x;
          printf "%s" (tl class)
      end
    fun fvtail (x,_) = case x of Typ => printf "" | _ => printf "()"
    fun fvtaili (x,_) = case x of Typ => printf "" | _ => printf "()"
  
    fun fnum 0 s1 s2 = printf "%s" s2
      | fnum i s1 s2 = s1 i
    fun fopt None _ g = g()
      | fopt (Some t) f _ = f t
    fun fstropt x = fopt x (fn s => fv(Str, s)) (fn _ => printf "")
    fun fsigopt x = fopt x (fn s => fv(Sig, s)) (fn _ => printf "Ÿ")
    fun ffctopt x = fopt x (fn s => fv(Fct, s)) (fn _ => printf "Լ") 
  
(*    fun pt _ = pp_ty *)
    fun ps _ = pp_scheme 
    fun pv _ = fv and pvtail _ = fvtail and pvtaili _ = fvtaili
    fun pstropt _ = fstropt and psigopt _ = fsigopt and pfctopt _ = ffctopt
    fun pstrtaili _ None = printf "" | pstrtaili _ _ = printf "()"
    fun f (Unbound s) = 
            printf "%a%a 𸨴ϴ." pv s pvtail s
      | f (Duplicate s) = 
            printf "%a%a ߺ Ǿϴ." pv s pvtaili s
      | f (UnifyFail(t1,t2)) = (
            TyPrint.start();
			printf "@[@[";
            TyPrint.typ' t1; 
            printf "@]@ Ÿ̾@ ϴµ@]@ @[@[";
            TyPrint.typ' t2;
            printf "@]@ ŸԴϴ.@]";
            TyPrint.ending())
    | f (TsApply(id,n,m)) = (
            printf "%a@ ڰ@ " pv (Typ,id);
            printf "%t ϴµ@ "
              (fn _ => fnum n (printf "%d ") "");
            printf "%tϴ."
              (fn _ => fnum m (printf "%d ") ""))
    | f (RecursiveType [a,b]) = (
        fv(Typ, [a]);
        printf "@ ";
        fv(Typ, [b]);
        printf "@ ǰ@  ¹ ֽϴ.")
    | f (RecursiveType ids) = (
          printf " ŸԵ ǰ@  ¹ ֽϴ:@ ";
          pp_list (fn s => printf "%s" s) (fn _ => printf ",@ ") ids;
          printf ".")
    | f (ConstraintSig id) = (
          fv(Typ, id);
          printf "@ ̹ Ÿ  ǵǾ@   ϴ.")
    | f (RConstraintSig id) = (
          fv(Typ, id);
          printf "@ ̹ ѵǾ@ ݺؼ   ϴ.")
    | f (FunName(id1,id2)) =
          printf "%s () ϴ @ %s () ߽ϴ." id1 id2
    | f (FunLength(n1,n2)) =
          printf "%d  ̾ ϴµ@ %d  Խϴ." n1 n2
    | f (Match(c,st,sg,(l,m))) = 
       let val s1 = case c of FS => "" | _ => ""
           val s2 = case c of FI => "ӹ" | _ => "Ÿ"
       in  (case c of
              SS => printf "%a%a@ %a@  ʽϴ.@;" pstropt st pstrtaili st psigopt sg
            | FI => printf "%a%a@ %a Է@ մϴ.@;" pstropt st pstrtaili st pfctopt sg
            | FS => printf "%a ǰ@ %a@  ʽϴ.@;" pfctopt st psigopt sg);
           (case m of
              Value(x,t,s) => printf "%a Ÿ@  ȵǱ Դϴ.@\n" pv (Var,l@[x]);
                              printf "@[<4>- %s@ %a ̰@]@\n@[<4>- %s@ %a Դϴ.@]" s1 ps t s2 ps s
            | Domain x => let val x' = add_str l x
                          in  printf "%a%a %s  Դϴ." pv x' pvtaili x' s1
                          end
            | Type(t,r) => 
                let val t' = (Typ, l@[t])
                in  case r of
                      Arity(i,j) => 
                        let fun f1 _ = fnum i (printf "%d ε") "µ"
                            fun f2 _ = fnum j (printf "%d ") ""
                        in  printf "%a ڰ %s@ %t@ %s@ %t Դϴ." pv t' s1 f1 s2 f2
                        end
                    | Constype(c,sc1,sc2) => 
                        printf "%a Ÿ  %s   Ÿ ٸ Դϴ.@\n" pv t' c;
                        printf "@[<4>- %s@ " s1;
                        fopt sc1 (printf "%a ̰" ps) (fn _ => printf "ڰ ");
                        printf "@]@\n@[<4>- %s@ " s2;
                        fopt sc2 (printf "%a Դϴ." ps) (fn _ => printf "ڰ ϴ")
                    | Consexist c =>
                        printf "%a@ Ÿ  %s @ %s@  Դϴ." pv t' c s2
                    | Consreq c =>
                        printf "%a@ Ÿ  %s @ %s@  Դϴ." pv t' c s1
                    | _ => 
                        printf "%a@ ٸ Դϴ." pv t'
                end)
       end
  in Location.print Format.std_formatter l; 
     printf "@[<6>:@ ";
     f t;
     printf "@]"
 end

 fun pp x = pp' x
 fun warning (x,l) = 
   let fun message Overload = " մϴ."
       val ppf = Format.std_formatter
       val msg = ": "^message x
       val newlines = ref 0
   (*    val _ = for i=0; i<String.length msg - 1; i+1 do
                 if msg.[i] = '\n' then incr newlines; *)
   in  fprintf ppf "%a" Location.print l;
       let val (out, flush, newline, space) =
              Format.pp_get_all_formatter_output_functions ppf ()
           fun countnewline x = (incr newlines; newline x)
       in  Format.pp_set_all_formatter_output_functions ppf out flush countnewline space;
           Format.fprintf ppf "%s@." msg;
           Format.pp_print_flush ppf ();
           Format.pp_set_all_formatter_output_functions ppf out flush newline space
       end;
       pp_print_flush ppf ();
       Location.incr_num_loc_lines !newlines
   end
end
