(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

functor HashFn (K: sig type t end) = struct
  structure K' = 
   struct type t = K.t
          fun equal (x:t) y = x=y
          val hash = Hashtbl.hash
   end
  open Hashtbl
  structure M = Make(K')
  type key = K.t
  type 'a t = 'a M.t
  fun empty() = M.create 30
  fun add m (k,_) v = (M.remove m k; M.add m k v)
  fun new (k,_) v = let val m = empty() in M.add m k v; m end
  fun find m (k,_) = M.find m k
  val iter = M.iter
  val mem = M.mem
  fun plus m1 m2 = M.iter (fn k v => (M.remove m1 k; M.add m1 k v)) m2
  fun exadd m (k,_) v = if M.mem m k then raise TyDebug.Bug "NEnv.HashFn.exadd"
                                     else M.add m k v
  fun explus m1 m2 = 
        M.iter (fn k v => if M.mem m1 k then raise TyDebug.Bug "NEnv.HashFn.explus"
                                        else M.add m1 k v) m2
  fun map f m = let val n = empty()
                in  M.iter (fn k v => M.add n k (f v)) m; n end

  fun to_list ve = 
    let val l = ref []
    in  M.iter (fn k v => l := (k,v)::(!l)) ve;
        !l
    end
end
 
structure NEnv = struct

  open Format
  open Info.StringInfo
  open String_ast.Ast
  open Ty

  structure R = TyError

  fun pp_list f g [] = ()
    | pp_list f g [h] = f h
    | pp_list f g (h::t)= (f h; g h; pp_list f g t)
(*
  fun pp_tns l = (printf "["; pp_list Ty.Print.tyname (fn _ => printf ", ") l; printf "]")
 *)
  structure VE = HashFn(struct type t = varidinfo end)
  structure TE = HashFn(struct type t = tyidinfo end)
  structure SE = HashFn(struct type t = stridinfo end)

  type ve = Ty.scheme VE.t
   and te = TS.t TE.t
   and se = env SE.t
   and env = Env of se * te * ve

  fun clos_ve c ve = VE.map (fn v => Ty.clos c (Ty.in_type v)) ve

  fun unify_ve a m1 m2 = (
    VE.iter (fn k v => Unify.it a (Ty.in_type v) (Ty.in_type (VE.M.find m2 k))) m1;
    VE.iter (fn k _ => ignore(VE.M.find m1 k)) m2) 
      handle Not_found => raise TyDebug.Bug "Env.ve_unify"
  

  fun tynames_te m =
    let val tns = ref Tns.empty
    in  TE.iter (fn _ ts => tns := Tns.union (!tns) (TS.tynames ts)) m;
        !tns
    end
  fun tynames_se se = 
    let val tns = ref Tns.empty
    in  SE.iter (fn _ e => tns := Tns.union (!tns) (tynames_env e)) se;
        !tns
    end
  and tynames_env(Env(se,te,_)) = Tns.union (tynames_se se) (tynames_te te)


structure E = struct
  type t = env

  fun empty() = Env(SE.empty(),TE.empty(),VE.empty())
  fun ve (Env(_,_,v)) = v
  fun te (Env(_,v,_)) = v
  fun se (Env(v,_,_)) = v
  fun from_ve ve = Env(SE.empty(),TE.empty(),ve)
  fun from_vete ve te = Env(SE.empty(),te,ve)
  fun from_se se = Env(se,TE.empty(),VE.empty())
  fun plus (Env(s1,t1,v1)) (Env(s2,t2,v2)) = 
        (SE.plus s1 s2; TE.plus t1 t2; VE.plus v1 v2)
  val tynames = tynames_env
end
 
structure Tnm = MapFn(Tn)

structure TNSubst = struct
  type t = Tn.t Tnm.t
  open TS
  fun apply_tau s tau = 
   let fun f tau = case tau of
           Link t => f (!t)
         | Fun(t1,t2) => Fun(f t1,f t2)
         | Rec x => (case Row.flatten x of 
                       (m,_) => Rec(List.map (fn (l,t)=>(l,f t)) m,None))
         | Var _ => tau
         | Data(l,tn) => let val l' = List.map f l
                         in (Data(l',Tnm.find tn s))
                               handle Not_found => Data(l',tn)
                         end 
         | Tuple l => Tuple(List.map f l)
   in  f tau
   end
  and apply_cl s cl = List.map (fn (l,None) => (l,None) 
                                 | (l,Some t) => (l,Some(apply_tau s t))) cl
  fun apply_scheme s (i,tau) = (i,apply_tau s tau)
  fun apply_ve s ve = VE.map (apply_scheme s) ve
  fun apply_ts s ts = case ts of
      TsName tn => (let val tn'= Tnm.find tn s 
                    in  (case CE.find tn of 
                           [] => () 
                         | l => CE.add tn' (apply_cl s l));
                        TsName tn'
                    end handle Not_found => ts)
    | TsFun(i,tau) => TsFun(i,apply_tau s tau)
  fun apply_te s te = TE.map (apply_ts s) te
  fun apply_se s se = SE.map (apply_e s) se
  and apply_e s (Env(se,te,ve)) = 
    let val te' = apply_te s te
        val ve' = apply_ve s ve
        val se' = apply_se s se
    in  Env(se',te',ve') end
end

structure TSubst = struct
  type t = TS.t Tnm.t

  open TS
  exception Invalid_apply

  val empty = Tnm.empty
  val add = Tnm.add
  val mem = Tnm.mem
(*
  fun pp s = Tnm.iter (fn k v => Ty.Print.tyname k; printf " -> "; TS.pp v; printf " ") s
*) 
  fun apply_tau s tau = 
   let fun f tau = case tau of
           Link t => f (!t)
         | Fun(t1,t2) => Fun(f t1,f t2)
         | Rec x => (case Row.flatten x of 
                       (m,_) => Rec(List.map (fn (l,t)=>(l,f t)) m,None))
         | Var _ => tau
         | Data(l,tn) => let val l' = List.map f l
                         in  (TS.apply l' (Tnm.find tn s))
                                 handle Not_found => Data(l',tn)
                         end 
         | Tuple l => Tuple(List.map f l)
   in  f tau
   end
  and apply_cl s cl = List.map (fn (l,None) => (l,None)
                                 | (l,Some t) => (l,Some(apply_tau s t))) cl
  fun apply_scheme s (i,tau) = (i,apply_tau s tau)
  fun apply_ve s ve = VE.map (apply_scheme s) ve
  fun apply_ts s ts = case ts of
      TsName tn => ((case Tnm.find tn s of
                          TsName x => ( 
                            (case CE.find tn of 
                               [] => ()
                             | l => CE.add x (apply_cl s l));
                            TsName x)
                        | x => if CE.find tn = [] then x else raise Invalid_apply
                    ) handle Not_found => ts)
    | TsFun(i,tau) => TsFun(i,apply_tau s tau)
  fun apply_te s te = TE.map (apply_ts s) te
  fun apply_se s se = SE.map (apply_e s) se
  and apply_e s (Env(se,te,ve)) = 
    let val te' = (apply_te s te)
        val ve' = (apply_ve s ve)
        val se' = (apply_se s se)
    in  Env(se',te',ve') end

end
 
structure G = struct

  type t = Tns.t * E.t
  type s = t
  fun from_env c e = (List.filter (fn (_,i,_,_) => i>c) (E.tynames e),e)
  fun to_env(_,e) = e
  fun t(s,_) = s

  fun make_new (l,e) c = 
   let val l' = List.map (Tn.copy c) l
       val subst = List.fold_right2 (fn t t' => Tnm.add t t') 
                     l l' Tnm.empty
   in  (l',TNSubst.apply_e subst e) 
   end
  fun remove_t (t1,e1) (t2,e2) = (Tns.diff t1 t2,e1)

  structure M = HashFn(struct type t = sigidinfo end)

  type m = s M.t
  val add = M.add val find = M.find


end

(* Functor and Functor Enviroment *)
structure F = struct

  type t = Tns.t * stridinfo list * E.t * G.t
  type s = t
  structure M = HashFn(struct type t = fctidinfo end)
  type m = s M.t
  val find = M.find val add = M.add

  fun make idl (t,e) sigma = (t,List.map (fn (x,_)=>x) idl,e,sigma)
  fun make_argument (_,l,_,_) sl = 
       let val info = Location.none
           (* for OCaml *)
           fun gen_strs 0 = [] 
             | gen_strs n = ("@fctarg:"^(string_of_int n))::gen_strs(n-1)
           val newstrs = gen_strs (List.length l)

       in  StrStr(SeqStrDec([
             StrDec(List.map2 (fn i s => ((i,info),None,s,info)) newstrs sl,info),
             StrDec(List.map2 (fn i s => ((i,info),None,VarStr(([],(s,info),info),info),info)) 
                                          l newstrs,info)]
             ,info),info)
       end
  fun env(_,_,_,g) = G.to_env g


end

structure Match = struct
  exception Error

  open TS
  fun check_bounds l = let
    fun f i [] = ()
      | f i (h::t) = if h = Tv.bound i then f (i+1) t else raise Error
   in f 0 l end
  


  val ln = Location.none

  (* first step:
       for structure e1 and signature (t,e2),
       scan each type definition in order to:
       - find a map (tyvar -> tystr) 
       - check illegal type specification such as (struct type t = A | B end :
                                                   sig    type t = int   end)
               except for the case (TsFun,TsFun)
       - check TE and SE's domain
  *)
 
  fun make_phi t e1 e2 = 
    let val phi = ref Tnm.empty
        fun ts s1 s2 = 
          let val ns1 = TS.arity s1 val ns2 = TS.arity s2
          in  (if ns1 <> ns2 then raise R.TM(R.Arity(ns1,ns2)));
              case (TS.normalize s1,TS.normalize s2) of
                (TsName n1,TsName n2) => 
                  if CE.find n2 <> [] && CE.find n1 = [] then raise R.TM(R.Name(n1,n2)) else
                    let val n' = (Tnm.find n2 (!phi)) handle Not_found => Some(TsName n2)
                    in  case n' of
                          Some(TsName n3) => if n1<>n3 then raise R.TM(R.Name(n1,n3))
                        | None => phi := Tnm.add n2 (Some(TsName n1)) (!phi)
                        | _ => raise Error
                    end
              | (TsFun c1,TsName n) =>
                  if CE.find n <> [] then raise Error else 
                    let val n' = (Tnm.find n (!phi)) handle Not_found => raise Error
                    in  case n' of
                          Some(TsFun c2) => ()
                        | None => phi := Tnm.add n (Some s1) (!phi)
                        | _ => raise Error
                    end
              | (TsFun s1,TsFun s2) => ()
              | _ => raise Error
          end
        fun te te1 te2 = TE.M.iter (fn k v =>
          let val t = (TE.M.find te1 k) handle Not_found =>
                                          raise R.M([],R.Domain(R.Typ,[k]))
          in  (ts t v) 
                handle R.TM x => raise R.M([],R.Type(k,x))
                     | Error => raise R.M([],R.Type(k,R.String(t,v)))
          end) te2
        fun se se1 se2 = SE.M.iter (fn k v =>
          let val t = (SE.M.find se1 k) handle Not_found =>
                                          raise R.M([],R.Domain(R.Str,[k]))
          in  (env t v) handle R.M(a,b) => raise R.M(k::a,b)
          end) se2
        and env e1 e2 = (te (E.te e1) (E.te e2); se (E.se e1) (E.se e2))
    in  List.iter (fn tn => phi := Tnm.add tn None (!phi)) t;
        env e1 e2;
        Tnm.mapi (fn k (Some t) => t 
                   | _ _ => raise TyDebug.Bug "env.Match.make_phi") (!phi)
    end
 
  (* second step:
       for structure e1 and signature (t,e2),
       - check VE's domain
       - check VE's types
       - check type's constructors environment
       - check type function
  *)

  fun str_sig e1 (t,e2) = 
    let val phi = make_phi t e1 e2
        fun equal t1 t2 = if not (Ty.equal t1 t2) then raise Error
        fun ty phi a t1 t2 = 
          let val alen = Array.length a 
              fun f t1 t2 = (
                case (t1,t2) of
                    (Link t1',_) => (case !t1' of
                       Var(Ol i) => (case t2 of
                         Link t2' => f t1 (!t2')
                       | Data(l,n) => ((f t1 (TS.apply l (Tnm.find n phi)))
                           handle Not_found => 
                            Unify.it Location.none t1 (Data(l,n)))
                       | _ => raise Error) (* 2.003 *)
                     | t1'' => f t1'' t2)
                  | (_,Link t2') => f t1 (!t2')
                  | (Fun(t11,t12),Fun(t21,t22)) => (f t11 t21; f t12 t22)
                  | (Tuple l1,Tuple l2) => List.iter2 f l1 l2
                  | (Tuple _,Rec r) => f t1 (Row.enforce_tuple r)
                  | (Rec r,Tuple _) => f (Row.enforce_tuple r) t2
                  | (Rec r1,Rec r2) => (case (Row.flatten r1,Row.flatten r2) of
                      ((m1,None),(m2,None)) => 
                        Labm.iter2 (fn l1 t1 l2 t2 => if l1=l2 then f t1 t2
                                                               else raise Error) m1 m2
                    | _ => raise TyDebug.Bug "Env.Match.Rec")
                  | (_,Data(l,n)) => ((f t1 (TS.apply l (Tnm.find n phi))) 
                        handle 
                          Not_found => (
                            case t1 of 
                              Data(l1,n1) => (
                                if n1<>n then raise Error;
                                List.iter2 f l1 l)
                            | Var(Bound x) => g x t2
                            | _ => raise Error))
                  | (Var(Bound x),_) => g x t2
                  | _ => raise Error)   
              and g x t = 
                if x >= alen then raise Error else
                  (case a.[x] of Some t1' => equal t1' t
                               | None => a.[x] <- Some t)          
          in  (f t1 t2) handle _ => raise Error end
       fun equal_wrt phi (n1,t1) (n2,t2) = 
          if n1=n2 then ty phi (Array.init n1 (fn n=>Some(Tv.bound n))) t1 t2
                   else raise Error
       and less_wrt phi (n1,t1) (n2,t2) = ty phi (Array.make n1 None) t1 t2
       fun cl phi n1 n2 = 
         let val i = Tn.arity n1
             fun foreach [] [] f = ()
               | foreach [] ((c,_)::_) f = raise R.TM(R.Consreq c)
               | foreach ((c,_)::_) [] f = raise R.TM(R.Consexist c)
               | foreach (h1::t1) (h2::t2) f = (f h1 h2; foreach t1 t2 f)
             val ce1 = CE.find n1 val ce2 = CE.find n2
         in  if ce1<> [] && ce2<>[] then
               foreach (CE.find n1) (CE.find n2) (fn (c1,t1) (c2,t2) =>
                 if c1=c2 then case (t1,t2) of
                      (Some t1,Some t2) => 
                        ((equal_wrt phi (i,t1) (i,t2)) handle 
                           Error => raise R.TM(R.Constype(c1,Some(i,t1),Some(i,t2))))
                    | (None,None) => ()
                    | (Some t,None) => raise R.TM(R.Constype(c1,Some(i,t),None))
                    | (None,Some t) => raise R.TM(R.Constype(c1,None,Some(i,t)))
                 else if c1<c2 then raise R.TM(R.Consexist c1)
                 else raise R.TM(R.Consreq c2))
         end
       fun ts phi s1 s2 = case (TS.normalize s1,TS.normalize s2) of
            (TsName n1,TsName n2) =>  cl phi n1 n2
          | (TsFun s1,TsFun s2) => equal_wrt phi s1 s2
          | (TsFun(i1,t1),TsName n2) => 
              ((case Tnm.find n2 phi of
                  TsFun(i2,t2) => equal t1 t2
                | _ => raise Error) handle Not_found => raise Error)  
          | _ => ()

       fun ve ve1 ve2 = VE.M.iter (fn k v => 
          let val t = (VE.M.find ve1 k) 
                        handle Not_found => raise R.M([],R.Domain(R.Var,[k]))
          in  (less_wrt phi t v)
                handle Error => raise R.M([],R.Value(k,t,v))
          end) ve2
        fun te te1 te2 = TE.M.iter (fn k v => 
          let val t = TE.M.find te1 k
          in  (ts phi t v)
                handle R.TM x => raise R.M([],R.Type(k,x))
                     | Error => raise R.M([],R.Type(k,R.String(t,v)))
          end) te2
        fun se se1 se2 = SE.M.iter (fn k v => ((env (SE.M.find se1 k) v)
                                               handle R.M(a,b) => raise R.M(k::a,b))) se2
        and env e1 e2 = (
          te (E.te e1) (E.te e2);
          se (E.se e1) (E.se e2);
          ve (E.ve e1) (E.ve e2))
    in  env e1 e2;
        phi
    end 

  val eg = str_sig

  fun ef e (t1,_,e1,sigma) c =
    let val phi = str_sig e (t1,e1)
        val (_,e2') = G.make_new sigma c
    in  TSubst.apply_e phi e2'
    end

end

  type t0 = RealEnv of E.t
          | OpenEnv of E.t * (strid list * E.t) list
   and estack = t0 list ref
  type b = {fcts: F.m, sigs: G.m, env: estack}

  type t1 = (b ref) * (b ref) * estack
  structure Tvm = HashFn(struct type t = tyvarinfo end)
  type t2 = ty Tvm.t list

  fun empty() =  {fcts = F.M.empty(), sigs = G.M.empty(), env = ref [RealEnv(E.empty())]}
  fun e_to_b e = {fcts = F.M.empty(), sigs = G.M.empty(), env = ref [RealEnv e]} 

  (* in order to keep context for each tyname *)
  val context = ref ([]: string list)
  fun set_context x = context := x
  fun push_context x = context := x::(!context)
  fun pop_context() = case !context of h::t => context := t 
                                     | _ => raise TyDebug.Bug "Env.pop_context"
  val sig_context_prefix = ref "*"
  fun sig_context' x = !sig_context_prefix^x
  fun file_context x = "%"^x
  fun top_context () = []
  val base_context = ref ([]: string list)
  fun reset_context() = context := !base_context
  fun functor_context () = "@"
  fun sig_context (x,_) = sig_context'(x)
  fun fct_context' x = "@"^x
  fun fct_context (x,_) = fct_context' x

  val (s as (s1,s2,s3)) = (ref(empty()), ref(empty()), ref []):t1
  val tvs = ref ([]:t2)

  fun get_b() = !s2
  fun set_b b = s2 := b
  fun get_base() = !s1
  fun set_base b = s1 := b

  fun reset c = (s2 := empty(); s3 := []; base_context := c; set_context c;
                 Tv.Ol.m := []; Row.m := [])
  fun reset_top () = reset (top_context())
  fun reset_file name = reset ([file_context name])

  fun push_empty () = s3 := RealEnv(E.empty())::(!s3)
  fun push_e e = s3 := RealEnv(e)::(!s3)
  fun push_ve ve = push_e (E.from_ve ve)

  exception Reduce
  fun reduce() = 
    let fun f (OpenEnv(e,_)::t) = let val e' = f t in E.plus e' e; e' end
          | f (RealEnv(e)::t) = (s3 := RealEnv(e)::t; e)
          | f [] = raise Reduce
    in  (ignore(f (!s3))) handle Reduce => ()
    end 
  fun pop() = (reduce();
               case !s3 of RealEnv(h)::t => (s3 := t; h)
                         | _ => raise TyDebug.Bug "Env.pop")

  fun top() = case !s3 of RealEnv(e)::t | OpenEnv(e,_)::t => e | [] => 
             (case !((!s2).env) of RealEnv(e)::t | OpenEnv(e,_)::t => e
                                 | [] => raise TyDebug.Bug "Env.top")

  fun get_e() = top()
  fun get_ve() = E.ve(get_e())
  fun get_te() = E.te(get_e())
  fun get_se() = E.se(get_e())
  fun get_g() = (get_b()).sigs
  fun get_f() = (get_b()).fcts

  fun add_vei k v = VE.add (get_ve()) k v
  fun add_ve ve = VE.M.iter (VE.M.add (get_ve())) ve
  fun add_tei k v = TE.add (get_te()) k v
  fun add_e e = (E.plus (get_e()) e)
  fun add_sei k v = SE.add (get_se()) k v
  fun add_sig k v = G.add (get_g()) k v
  fun add_f k a b c = F.add (get_f()) k (F.make a b c)

  fun add_b' b (f,g,s) = (F.M.plus b.fcts f;
                          G.M.plus b.sigs g;
                          b.env := RealEnv(E.from_se s) :: !(b.env))

  fun add_b data = add_b' (get_b()) data
  fun add_base data = add_b' (get_base()) data
  fun move_b_to_base() = 
    let val x1 = !s1
        val x2 = !s2
    in  F.M.plus x1.fcts x2.fcts;
        G.M.plus x1.sigs x2.sigs;
        x1.env := !(x2.env) @ !(x1.env);
        s2 := empty()
    end

  fun push_tvs tvl = let val m = Tvm.empty() in 
        List.iter (fn (x as (tv,_)) => Tvm.add m x (Tv.new_named tv)) tvl;
        tvs := m::(!tvs) 
       end 
  fun push_tvs_bound tvl = let val m = Tvm.empty() val i = ref (-1) in
        List.iter (fn tv => Tvm.add m tv (i++; Tv.bound (!i))) tvl;
        tvs := m::(!tvs)
       end
  fun pop_tvs() = case !tvs of (h::t) => tvs := t 
                             | _ => raise TyDebug.Bug "Env.S.pop_tvs"
  fun find_tv id = (let
    fun f [] = raise Not_found
      | f (h::t) = (Tvm.find h id) handle Not_found => f t
   in f (!tvs) end) handle Not_found => R.unbound_tyvar id

  open Format

  fun find_env' env [] = env
    | find_env' env sl = List.fold_left (fn (Env(se,_,_)) id => SE.find se id) env sl

  fun current_estack () = (!s3) @ (!(!s2.env)) @ (!(!s1.env))

  fun find_open (lid as ([],id,a)) f =
        let fun g [] = raise Not_found
              | g (RealEnv(e)::t) = ((f e id,lid) handle Not_found => g t)
              | g (OpenEnv(e,l)::t) =
                  let fun h nil = g t
                        | h ((s,e')::t') = ((f e' id, (s,id,a)) handle Not_found => h t')
                  in  h (([],e)::l)
                  end
        in  g (current_estack())
        end
    | find_open (lid as (l,id,a)) f = 
        let fun g [] = raise Not_found
              | g (RealEnv(e)::t) =
                       let val e' = (Some(find_env' e l)) handle Not_found => None
                       in  case e' of None => g t
                                    | Some e'' => (f e'' id, lid)
                       end
              | g (OpenEnv(e,l')::t) =
                       let fun h nil = g t
                             | h ((s,e')::t') =
                                let (* val _ = printf "Look %s for " ((fn(x,_)=>x) id);
                                            List.iter (fn(s',_) => printf "%s." s') s;
                                            printf "\n" *)
                                    val eo = (Some(find_env' e' l)) 
                                               handle Not_found => None
                                in  case eo of None => h t'
                                             | Some e'' => (f e'' id, (s@l,id,a))
                                end
                       in  h (([],e)::l')  
                       end
        in  g (current_estack())
        end

  fun find_in_b f = (f (!s2)) handle Not_found => (f (!s1))
 
  fun find_ty lid = (find_open lid (fn e => VE.find (E.ve e)))
                       handle Not_found => R.unbound_var lid
  fun find_tyfn lid = (find_open lid (fn e => TE.find (E.te e)))
                         handle Not_found => R.unbound_ty lid
  fun find_e lid = (find_open lid (fn e => SE.find (E.se e)))
                      handle Not_found => R.unbound_str lid

  fun find_tyfn_in e (sl,id,a) = (TE.find (E.te(find_env' e sl)) id)
                           handle Not_found => R.unbound_ty (sl,id,a)

  fun find_sig id = (find_in_b (fn b => G.find b.sigs id))
                       handle Not_found => R.unbound_sig id
  fun find_f id = (find_in_b (fn b => F.find b.fcts id)) 
                       handle Not_found => R.unbound_fct id

  fun add_open sl = 
    let val l0 = List.map (fn id => let val (x,(l,i,a))=find_e id in ((l,i,a),x) end) sl
        val l1 = List.map (fn ((l,i,a),x) => (l@[i],x)) l0
        val l2 = List.map (fn (x,y) => x) l0
    in  s3 := OpenEnv(E.empty(), l1)::(!s3);
        l2
    end

  fun init_open s = 
    (let val s' = (s,Location.none)
        val (x,_) = find_e ([],s',Location.none)
     in  !s1.env := OpenEnv(E.empty(), [([s'],x)])::!(!s1.env)
     end) handle _ => ()

  (* To compile .ns files.
     assume that s2 has only signatures *)

  fun convert_to_str() = 
    let val g = (!s2).sigs
        val b = empty()
    in  case !(b.env) of 
          [RealEnv(Env(se,_,_))] => ( 
            G.M.iter (fn k (t,e) => SE.M.add se k e) g;
            s2 := b)
        | _ => raise TyDebug.Bug "Env.convert_to_str" 
    end

  fun stabilize_top() = (!s2.env := !s3 @ !(!s2.env); s3 := []) 
  fun stabilize_file() =
    let fun f l = case l of
            [RealEnv e] => (e,[])
          | OpenEnv(e,_)::t | RealEnv(e)::t => let val (e',l) = f t in (e',l@[e]) end
          | _ => raise TyDebug.Bug "Env.stabilize_file"
        val (e,el) = f (!s3 @ !(!s2.env))
    in  List.iter (E.plus e) el;
        s3 := []; !s2.env := [RealEnv e]; e
    end
end



 
