(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(* ---------------- Predef ----------------- *)

structure Predef = struct

  open Ty
  open NEnv

  val type_int =    Ty.type_int
  val type_real =   Ty.type_real   
  val type_string = Ty.type_string
  val type_bool =   Ty.type_bool
  val type_unit =   Ty.type_unit
  val type_char =   Ty.type_char
  val type_exn =    Ty.type_exn

  val gen_ref =   Ty.gen_ref
  val gen_list =  Ty.gen_list
  val gen_array = Ty.gen_array 
  (* for OCAML *) val gen_option = Ty.gen_option

  val te = let val alpha = Tv.new() val te = TE.empty() in
    List.iter
     (fn t => (
       case t of Data(_,t) => TE.M.add te (Tn.name t) (TS.gen_name t) 
               | _ => raise TyDebug.Bug "Predef.make the pervasive TE"))
     [type_int, type_real, type_string, type_bool, type_unit, type_char,
      type_exn, gen_ref alpha, gen_array alpha, gen_list alpha, 
      (* for OCAML *) gen_option alpha]; 
    te
   end
  fun three t1 t2 t3 = gen_fun t1 (gen_fun t2 t3)
  fun biop t = three t t t
  fun unop t = gen_fun t t
  val irbound = Var(Ol 0)
  val ve = let
    val ve = VE.empty()
  in
    List.iter (fn (l,t) =>
      List.iter (fn x => VE.M.add ve x t) l) 

  [(["+","-","*","/","**"], (1,biop irbound)),
   (["unary_minus"], (1,unop irbound)),
   (["%","<<",">>"], (0,biop type_int)),
   (["++","--"], (1,gen_fun (gen_ref irbound) type_unit)),
   (["+=","-=","*=","/="], (1,let val a=irbound in three (gen_ref a) a type_unit end)),
   (["true","false"], (0,type_bool)),
   (["True","False"], (0,type_bool)),
   (["andalso","&&","orelse","||"], (0,biop type_bool)),
   (["not"], (0,unop type_bool)),
   (["=","<>","=="], (1,let val a=Tv.bound 0 in three a a type_bool end)),
   (["<",">","<=",">="], (1,let val a=Tv.bound 0 in three a a type_bool end)),
   (["nil"], (1,gen_list(Tv.bound 0))),
   (["::"], (1,let val a=Tv.bound 0 val al=gen_list a in gen_fun (gen_pair a al) al end)),
   (["@"], (1,let val a=gen_list(Tv.bound 0) in biop a end)),
   (["^"], (0,biop type_string)),
   (["Match","Zero","Overflow","Bound","Equality"], (0,type_exn)),
   ([":="], (1,let val a=Tv.bound 0 in gen_fun (gen_ref a) (gen_fun a type_unit) end)),

(* for OCAML *)
   (["Not_found","Out_of_memory","Stack_Overflow","End_of_file","Division_by_zero",
     "Sys_blocked_io"], (0,type_exn)),
   (["Invalid_argument","Failure","Sys_error"], (0, gen_fun type_string type_exn)),
   (["Match_failure","Assert_failure"], 
     (0, gen_fun (Ty.Tuple[type_string, type_int, type_int]) type_exn)),
   (["None"],(1,let val a=Tv.bound 0 in gen_option a end)),
   (["Some"],(1,let val a=Tv.bound 0 in gen_fun a (gen_option a) end))

  ];    
  ve end

  val init = e_to_b(E.from_vete ve te)
end

