(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

structure EnvPrint = struct
  open Format
  open Ty
  open NEnv

  fun separate() = printf "@\n"
  fun none _ = ()

  fun print_map sw m g h f = (* switch map seperator iterator element-function *)
        h (fn x y => f x y (fn x=>if !sw then g() else sw := true)) m
  fun print_list sw m g f = (* switch map seperator element-function *)
        List.iter (fn x => f x (fn x=>if !sw then g() else sw := true)) m

  fun tystr ts c id = case ts of
        TS.TsName x =>
          if Tn.context x <> c || Tn.name x <> id then 
            (printf "@ =@ "; TyPrint.tynamec x c)
          else 
            let val cl = CE.find x
            in  case cl of [] => () 
                         | _ => let val sw = ref false
                                in  printf "@ =@ @[<-2>";
                                    print_list sw cl (fn _ => printf "@;| ")
                                      (fn (i,None) g => g(); printf "%s" i
                                        | (i,Some t) g => g(); printf "@[<2>%s of@ " i;
                                                          TyPrint.typc t c;
                                                          printf "@]");
                                    printf "@]"
                                end
            end
      | TS.TsFun(n,t) => printf "@ =@ @["; TyPrint.typc t c; printf "@]"

  fun printable_ve x = x < "A" || x > "Z"
  fun printable_te x = true
  fun printable_se x = true
  fun count_printable (Env(s,t,v)) =
    let val counter = ref 0
    in  SE.iter (fn k _ => if printable_se k then counter++) s;
        TE.iter (fn k _ => if printable_te k then counter++) t;
        VE.iter (fn k _ => if printable_ve k then counter++) v;
        !counter
    end

  fun ve m c sw print_val = print_map sw m separate VE.iter 
    (fn id (n,t) g => let val x = String.sub id 0 1
                      in  if printable_ve id then 
                          (g();
                           printf "val %s@[: @[" id;
                           TyPrint.typc t c;
                           printf "@]";
                           print_val id;
                           printf "@]")
                          else if is_exn t then 
                          (g();
                           printf "@[exception %s" id;
                           (case exn_arg t of
                              Some t => (printf " of @["; TyPrint.typc t c; printf "@]")
                            | _ => ());
                           printf "@]")
                      end)
                    

  fun te m c sw = print_map sw m separate TE.iter
    (fn id e g => if printable_te id then (
                    g();
                    printf "@[<5>type %s" id;
                    tystr e c id;
                    printf "@]";
                    sw := true))

  fun e(env as Env(s,t,v)) c f = 
    let val sw = ref false
        val items = count_printable env
    in  printf "@[@[<2>sig";
        if items=1 then printf "@ " else if items>1 then printf "@\n";
        te t c sw;
        se s c sw;
        ve v c sw f;
        if items<=1 then printf "@]@ " else printf "@]@\n";
        printf "end@]"
    end

  and se m c sw = print_map sw m separate SE.iter
    (fn id env g => if printable_se id then ( 
           g();
           printf "@[<2>structure %s:@ " id;
           e env (id::c) none;
           printf "@]"))

  and print_g m sw = G.M.iter (fn k (t,e') => 
                       if !sw then printf "@\n" else sw := true;
                       printf "@[<2>signature %s =@ " k;
                       e e' (sig_context' k :: !base_context) none;
                       printf "@]") m

  fun print_f m sw = 
    let
      fun print_sig'(t,e') = case e' of Env(s,_,_) =>
            let val sw = ref false
            in  se s (functor_context()::!base_context) sw
            end
    in  
        F.M.iter (fn k (t,l,e1,(t',e2)) => 
                  (if !sw then printf "@\n" else sw := true;
                   printf "@[<2>functor %s@ (@[" k;
                   (case l of [sid] => (printf "%s:@ " sid; 
                                        e e1 (sid :: !base_context) none)
                            | _ => print_sig' (t,e1));
                   printf "@]) ->@ ";
                   e e2 (fct_context' k :: !base_context) none; 
                   printf "@]")) m
    end

  fun estack l g sw = 
    let val c = !base_context
        fun f l = case l of
            NEnv.RealEnv(Env(s,t,v))::tail => (
              te t c sw;
              se s c sw;
              ve v c sw g;
              f tail)
          | NEnv.OpenEnv(e',l)::t => f t
          | _ => ()
    in  f l
    end
   
  fun b (a:NEnv.b) h = 
        let val sw = ref false
        in  printf "@["; 
            print_g a.sigs sw; 
            print_f a.fcts sw; 
            estack (!(a.env)) h sw; 
            printf "@."
        end
end
