(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

structure TyPrint = struct
  open Ty
  open Format

  type ty = Ty.ty
  structure I = MakeMap(struct type t = int val compare = compare end)
  val count = ref 0
  val m = ref I.empty
  val bound_str = ref "'"
  val meta_str = ref "_"

  fun reset() = (count := 0; m := I.empty)

  fun close() = 
    let val sw = ref true
    in  I.iter (fn _ (_,false) => ()
                 | _ (i,true) =>  (if !sw then (printf "@ ("; sw := false)
                                          else printf ",@ ";
                                   printf "%s = " (!meta_str ^ Tv.make_tvname i);
                                   printf "int | real"))
        (!m);
        if not (!sw) then printf ")"
    end

  fun meta x attop = 
    let val (num,_) = (I.find x (!m)) handle Not_found => 
                         let val num = !count
                         in  count := !count + 1;
                             m := I.add x (num,attop) (!m);
                             (num, attop)
                         end
    in  printf "%s" (!meta_str ^ Tv.make_tvname num) 
    end

  fun tyvar x = case x of
        Meta(x,y) => meta x false
      | Ol i => meta i true
      | Bound i => printf "%s" (!bound_str ^ Tv.make_tvname i)
      | Name(_,s) => printf "[%s]" s

  fun pp_list f g nil = ()
    | pp_list f g (h::nil) = f h
    | pp_list f g (h::t) = (f h; g h; pp_list f g t)

  fun tyname (s,i,x,l) = 
    let fun f [] = ()
          | f [s] = printf "%s" s
          | f ["Pervasives",s] = printf "%s" s
          | f (h::t) = (if String.sub h 0 1 >= "A" then printf "%s." h; 
                        f t)
    in  f (List.rev(s::l))
    end



  fun tynamec (s,i,x,l) c =
    let fun f (h1::t1) (h2::t2) = if h1=h2 then f t1 t2 else tyname(s,i,x,List.rev(h1::t1))
          | f l _ = tyname(s,i,x,List.rev l)
    in  f (List.rev l) (List.rev c)
    end

  fun typc' ty c = let
    fun f t = case t of
        Link t' => f (!t') (* (printf "|"; pp (!t'); printf "|") *)
      | Var v => tyvar v
      | Fun(t1,t2) => (
          (case t1 of Fun _ => (printf "@[("; f t1; printf ")@]")
                    | _ => f t1);
          printf "@ ->@ "; f t2)
      | Data(args,tn) => let
          fun complex t = case t of
              Link l => complex (!l)
            | Tuple _ => true
            | Fun _ => true
            | _ => false
          val par = case args of
                [] => false
              | [arg] => complex arg
              | _ => true
         in
          if par then printf "(";
          pp_list f (fn _ => printf ",@ ") args;
          if par then printf ")";
          if List.length args > 0 then printf "@ ";
          tynamec tn c
         end
      | Rec x => let val (m,r) = Row.flatten x in
          printf "@[<1>{";
          pp_list (fn(k,t) => (printf "%s: " k; f t)) (fn _ => printf ",@ ") 
                  (Labm.to_list m);
          (case r of Some _ => printf ",@ ..." | _ => ());
          printf "}@]"
         end
      | Tuple tuple => 
         let fun g t = case unlink t of
               Fun _ | Tuple _ => printf "@[("; f t; printf ")@]"
             | _ => f t
         in  pp_list g (fn _ => printf "@ * ") tuple
         end
    in  
     f ty   
    end
  
  fun typ' ty = typc' ty []
  fun start() = reset()
  fun ending() = close()
  fun typc ty c = (start(); typc' ty c; ending())
  fun typ ty = (start(); typ' ty; ending())
  fun type_scheme (n,t) = typ t

  val pp = typ
end

