(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

structure TyUtil = struct

  open Ty

  (* is t a function type? *)
  fun is_fun t = case t of
      Fun _ | Var _ => true
    | Link t' => is_fun (!t')
    | _ => false

  (* is t a tuple type? *)
  fun is_tuple t = case t of
      Tuple _ => true
    | Link t' => is_tuple (!t')
    | Rec r => 
	    (case Row.flatten r of 
		   (m,None) => ((ignore(Row.labm2tuple m); true) handle _ => false)
	     | _ => raise Debug.CompilerBug "TyUtil.is_tuple.Rec")
    | _ => false

  (* does t has a function type? *)
  fun has_fun t = case t of
      Fun _ | Var _ => true
    | Tuple a | Data(a,_) => List.exists has_fun a
    | Rec r => (case Row.flatten r of (m,None) => List.exists (fn (l,t) => has_fun t) m
                                    | _ => raise Debug.CompilerBug "TyUtil.has_fun.Row")
    | Link t' => has_fun (!t')

  (* split_type: ty -> int -> ty * ty *)
  (* Input:  a function type and a number of arguments *)
  (* Output: a tuple of input types and an output type *)
  fun split_type t cnt = 
    let fun f t l 0 = (Tuple l,t)
          | f t l n = case t of Fun(t1,t2) => f t2 (l@[t1]) (n-1)
                              | Link t' => f (!t') l n
                              | _ => raise Debug.CompilerBug "Ty.split_type"
    in  f t [] cnt 
    end

  (* app_type: ty -> int -> ty *)
  fun app_type t cnt = 
    let fun f t cnt = case t of
            Fun(t1,t2) => if cnt=1 then t2 else f t2 (cnt-1)
          | Link t' => f (!t') cnt
          | _ => raise Debug.CompilerBug "TyUtil.app_type"
    in  if cnt < 1 then raise Debug.CompilerBug "TyUtil.app_type"
                   else f t cnt
    end

  fun is_exn (Link t') = is_exn (!t')
    | is_exn (Data([],("exn",0,0,[]))) = true
    | is_exn _ = false

  (* is_exn_type: ty -> bool *)
  fun is_exn_type t = case t of
        Var _ => true
      | Fun(_,t) => is_exn t
      | Link t => is_exn_type (!t)
      | _ => is_exn t

  fun clear_type t = clos 0 t

  fun unifiable t1 t2 = 
    let val l = Location.none
        val t1' = instance l (clos 0 t1)
        val t2' = instance l (clos 0 t2)
    in  (Unify.it l t1' t2'; true) handle _ => false
    end

  fun print_ty t = TyPrint.typ t
end
