(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

structure Unify = struct
 open Format
 open Ty
 exception Fail
 exception Occur

 fun unify_fail l t1 t2 = raise TyError.T(TyError.UnifyFail(t1,t2),l)

 fun occur v t = let
   fun f t = case t of
       Link t => if (==) v t then raise Occur else f (!t)
     | Fun(t1,t2) => (f t1; f t2)
     | Tuple l | Data(l,_) => List.iter f l
     | Rec r => frow (Row r)
     | Var _ => ()
   and frow r = case r of
       Rvar _ => ()
     | Row(m,None) => Labm.M.iter (fn _ => f) m
     | Row(m,Some r) => (Labm.M.iter (fn _ => f) m; frow !r)
  in f t end

 fun lowering c t = let
   fun f t = case t of
      Link t' => (case !t' of 
        Var(Meta(x,y)) => if y>c then t' := Var(Meta(x,c)) else ()
      | t => f t)
    | Data(l,_) | Tuple l => List.iter f l
    | Fun(t1,t2) => (f t1; f t2)
    | Rec x => (case Row.flatten x of (m,x) => Labm.M.iter (fn k v => f v) m)
    | _ => ()
  in f t end

 fun rv_level v = case !v of Rvar(_,x) => x
                           | _ => raise TyDebug.Bug "Unify.rv_level"

 and lowering_by_rv v m =
   let val lv = rv_level v
   in  Labm.M.iter (fn _ => lowering lv) m
   end 

   
 fun unlink' x = 
   let fun f x = case x of
         Link t => (case !t of Var(Meta _) | Var(Ol _) => x 
                             | x => let val x' = unlink x in t := x'; x' end)
       | _ => x
   in  f x 
   end
 
 fun it l t1 t2 = let 
   fun f t1 t2 = case (t1,t2) of 
       (Link _,_) | (_,Link _) => 
        let val t1 = unlink' t1 val t2 = unlink' t2
        in  case (t1,t2) of
              (Link t3,Link t4) => (if (==) t3 t4 then () else 
                case (!t3,!t4) of
                  (Var(Meta(_,x)),Var(Meta(_,y))) => if x<y then t4 := t1 else t3 := t2
                | (Var(Ol _),Var(Meta _)) 
                | (Var(Ol _),Var(Ol _)) => t4 := t1
                | (Var(Meta _),Var(Ol _)) => t3 := t2
                | _ => raise TyDebug.Bug "Unify.it.Link")
            | (Link a,b) | (b,Link a) => (
                case (!a,b) of
                  (Var(Meta(_,x)),_) => (occur a b; lowering x b; a := b)
                | (Var(Ol _),Data([],(n,0,0,[]))) => (
                    case n of
                      "int" | "real" => a := b
                    | _ => raise Fail)
                | _ => raise Fail) 
            | _ => f t1 t2
        end
     | (Data(l1,n1),Data(l2,n2)) => if n1=n2 then fs l1 l2 else raise Fail
     | (Fun(t3,t4),Fun(t5,t6)) => (f t3 t5; f t4 t6)
     | (Tuple a,Tuple b) => fs a b
     | (Rec x,Tuple a) | (Tuple a,Rec x)=> f (Rec x) (Row.of_tuple a)
     | (Rec r1,Rec r2) => 
        let val (m1,r1) = Row.flatten r1
            val (m2,r2) = Row.flatten r2
        in  case (r1,r2) of
              (None,None) => flabm m1 m2
            | (None,Some l2) => 
              let val ((c1,c2),(b1,b2)) = Labm.split m1 m2
              in  if Labm.is_empty b2 then (
                     flabm c1 c2; 
                     lowering_by_rv l2 b1;
                     l2 := Row(b1,None))
                  else raise Fail         
              end 
            | (Some l1,None) => f t2 t1
            | (Some l1,Some l2) =>
              let val ((c1,c2),(b1,b2)) = Labm.split m1 m2
                  val lv1 = rv_level l1 val lv2 = rv_level l2
                  val newlv = if lv1 > lv2 then lv2 else lv1
                  val newrv = Row.new_with_level newlv
              in  flabm c1 c2; 
                  lowering_by_rv l1 b2; lowering_by_rv l2 b1; 
                  l1 := Row(b2,newrv); l2 := Row(b1,newrv)
              end 
        end
     | (Rec _, Data([],("unit",0,0,[]))) => f t1 (Rec([],None))
     | (Data([],("unit",0,0,[])), Rec _) => f (Rec([],None)) t2
     | (Tuple[], Data([],("unit",0,0,[]))) | (Data([],("unit",0,0,[])), Tuple[]) => ()
     | (Var(Bound s),Var(Bound t)) => if s=t then () else raise Fail
     | (Var(Name s),Var(Name t)) => if s=t then () else raise Fail
     | _ => raise Fail 
   and fs l1 l2 = List.iter2 f l1 l2
   and flabm m1 m2 = 
         Labm.iter2 (fn l1 t1 l2 t2 => if l1=l2 then f t1 t2 else raise Fail) m1 m2
 in  
(*     printf "Unify on ";
     TyPrint.typ t1;
     printf " and ";
     TyPrint.typ t2;
     printf "\n"; *)
     (f t1 t2) handle (x as TyDebug.Bug _) => raise x
                    | _ => unify_fail l t1 t2
 end  

  (* Special unifier when c is a constant type
                            such as Data([],(name,0,0)) *)
  fun const a rho c = case unlink rho of
        (r as (Data _)) => if r<>c then unify_fail a r c 
      | _ => it a rho c

  (* Special unifier when v is a new type variable *)
  fun var t v = case v of Link l => l := t | _ => raise Fail

  (* Special unifier when t2 is a data type which has only a new type variable *)   
  fun data a t1 t2 = case t1 of
        Link t => (case !t of
          Var(Meta(_,x)) => (lowering x t2; t := t2)
        | t' => data a t' t2)
      | Data([s],n) => (case t2 of
          Data([s'],n') => if n=n' then ((var s s') handle _ => unify_fail a t1 t2)
                                   else unify_fail a t1 t2
        | _ => raise TyDebug.Bug "Unify.data")
      | _ => unify_fail a t1 t2

  (* Special unifier when tl are new type variables *)
  fun tuple a t tl = case t of
        Link t => (case !t of
          Var(Meta(_,x)) => (lowering x (Tuple tl); t := Tuple tl)
        | t' => tuple a t' tl)
      | Tuple l => 
	      if List.length l <> List.length tl then 
		    unify_fail a t (Tuple tl) 
	      else 
		    ((List.iter2 var l tl) handle _ => unify_fail a t (Tuple tl))
      | Rec r => ((it a t (Row.of_tuple tl)) handle _ => unify_fail a t (Tuple tl))
      | _ => unify_fail a t (Tuple tl)

  (* Special unifier when t2 is a new type variable *)
  (* Note that Unify.app makes an error message that 
     t3 is expected but t2->t3 is. *)
   
  fun app a b t1 t2 t3 = let val t4 = Fun(t2,t3) in case t1 of
        Link t => (case !t of
          Var(Meta(_,x)) => ((lowering x t4; occur t t3; t := t4)
                               handle _ => unify_fail b t4 t1)
                                            (* 2.002 *)
        | t' => app a b t' t2 t3)
      | Fun(t11,t12) => (var t11 t2; it a t3 t12)
      | _ => unify_fail b t4 t1 end

  (* Special unifier when t2 and t3 are new type variables *)
  fun lambda a t1 t2 t3 = let val t4 = Fun(t2,t3) in case t1 of
        Link t => (case !t of
          Var(Meta(_,x)) => (lowering x t4; t := t4)
       | t' => lambda a t' t2 t3)
      | Fun(t11,t12) => ((var t11 t2; var t12 t3) handle _ => unify_fail a t1 t4)
      | _ => unify_fail a t1 t4 end
end
